@@ -3,6 +3,7 @@
 
 #include "provider.h"
 #include "platform/endian.h"
+#include "openssl/rand.h"
 #include <string.h>
 
 #define DFLT_DIGEST "SHA256"
@@ -130,64 +131,13 @@ struct key_generator {
     void *cb_arg;
 };
 
-static void *p11prov_common_gen_init(void *provctx, int selection,
-                                     CK_KEY_TYPE type,
-                                     const OSSL_PARAM params[])
-{
-    struct key_generator *ctx = NULL;
-    /* big endian 65537 */
-    unsigned char def_e[] = { 0x01, 0x00, 0x01 };
-    int ret;
-
-    P11PROV_debug("common gen_init %p", provctx);
-
-    ret = p11prov_ctx_status(provctx);
-    if (ret != CKR_OK) {
-        return NULL;
-    }
-
-    if ((selection & OSSL_KEYMGMT_SELECT_KEYPAIR) == 0) {
-        P11PROV_raise(provctx, CKR_ARGUMENTS_BAD, "Unsupported selection");
-        return NULL;
-    }
-
-    ctx = OPENSSL_zalloc(sizeof(struct key_generator));
-    if (ctx == NULL) {
-        P11PROV_raise(provctx, CKR_HOST_MEMORY, "Failed to get key_generator");
-        return NULL;
-    }
-    ctx->provctx = (P11PROV_CTX *)provctx;
-    ctx->type = type;
-
-    /* set defaults */
-    switch (type) {
-    case CKK_RSA:
-        ctx->mechanism.mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN;
-        ctx->data.rsa.modulus_bits = 2048;
-        ctx->data.rsa.exponent_size = sizeof(def_e);
-        memcpy(ctx->data.rsa.exponent, def_e, ctx->data.rsa.exponent_size);
-        break;
-    case CKK_EC:
-        ctx->mechanism.mechanism = CKM_EC_KEY_PAIR_GEN;
-        ctx->data.ec.ec_params = prime256v1_param;
-        ctx->data.ec.ec_params_size = sizeof(prime256v1_param);
-        break;
-    case CKK_EC_EDWARDS:
-        ctx->mechanism.mechanism = CKM_EC_EDWARDS_KEY_PAIR_GEN;
-        break;
-    default:
-        P11PROV_raise(provctx, CKR_ARGUMENTS_BAD, "Invalid type %lu", type);
-        OPENSSL_free(ctx);
-        return NULL;
-    }
-
-    ret = p11prov_common_gen_set_params(ctx, params);
-    if (ret != RET_OSSL_OK) {
-        p11prov_common_gen_cleanup(ctx);
-        ctx = NULL;
-    }
-    return ctx;
-}
+const CK_MECHANISM_TYPE p11prov_rsapss_mechs[P11PROV_N_RSAPSS_MECHS] = {
+    CKM_SHA1_RSA_PKCS_PSS,     CKM_SHA224_RSA_PKCS_PSS,
+    CKM_SHA256_RSA_PKCS_PSS,   CKM_SHA384_RSA_PKCS_PSS,
+    CKM_SHA512_RSA_PKCS_PSS,   CKM_SHA3_224_RSA_PKCS_PSS,
+    CKM_SHA3_256_RSA_PKCS_PSS, CKM_SHA3_384_RSA_PKCS_PSS,
+    CKM_SHA3_512_RSA_PKCS_PSS, CKM_RSA_PKCS_PSS
+};
 
 static int p11prov_common_gen_set_params(void *genctx,
                                          const OSSL_PARAM params[])
@@ -452,11 +402,11 @@ static CK_RV common_key_usage_to_tmpl(struct key_generator *ctx,
     return CKR_OK;
 }
 
-static void *p11prov_common_gen(struct key_generator *ctx,
-                                CK_ATTRIBUTE *pubkey_template,
-                                CK_ATTRIBUTE *privkey_template, int pubtsize,
-                                int privtsize, OSSL_CALLBACK *cb_fn,
-                                void *cb_arg)
+static int p11prov_common_gen(struct key_generator *ctx,
+                              CK_ATTRIBUTE *pubkey_template,
+                              CK_ATTRIBUTE *privkey_template, int pubtsize,
+                              int privtsize, OSSL_CALLBACK *cb_fn, void *cb_arg,
+                              void **key)
 {
     CK_SLOT_ID slotid = CK_UNAVAILABLE_INFORMATION;
     CK_BYTE id[16];
@@ -474,7 +424,7 @@ static void *p11prov_common_gen(struct key_generator *ctx,
                                    pubtsize, privtsize);
     if (ret != CKR_OK) {
         P11PROV_raise(ctx->provctx, ret, "Failed to map Key Usage");
-        return NULL;
+        return ret;
     }
 
     if (ctx->uri) {
@@ -486,7 +436,7 @@ static void *p11prov_common_gen(struct key_generator *ctx,
                               ctx->mechanism.mechanism, NULL, NULL, true, true,
                               &session);
     if (ret != CKR_OK) {
-        return NULL;
+        return ret;
     }
 
     if (cb_fn) {
@@ -498,9 +448,13 @@ static void *p11prov_common_gen(struct key_generator *ctx,
     sh = p11prov_session_handle(session);
 
     if (cka_id.ulValueLen == 0) {
+        int err = RET_OSSL_ERR;
         /* generate unique id for the key */
-        ret = p11prov_GenerateRandom(ctx->provctx, sh, id, sizeof(id));
-        if (ret != CKR_OK) {
+        err = RAND_bytes_ex(p11prov_ctx_get_libctx(ctx->provctx), id,
+                            sizeof(id), 0);
+        if (err != RET_OSSL_OK) {
+            ret = CKR_GENERAL_ERROR;
+            P11PROV_raise(ctx->provctx, ret, "Failed to source random buffer");
             goto done;
         }
         cka_id.type = CKA_ID;
@@ -535,7 +489,12 @@ static void *p11prov_common_gen(struct key_generator *ctx,
         goto done;
     }
 
-    ret = p11prov_merge_pub_attrs_into_priv(pub_key, priv_key);
+    /* set the public key object as associated object of the private key,
+     * this way a public key can always be found from the private key and
+     * operations that assume an EVP_PKEY represent both can find what
+     * they need.
+     * This operation takes a reference so we can safely free pub_key */
+    p11prov_obj_set_associated(priv_key, pub_key);
 
 done:
     if (ret != CKR_OK) {
@@ -544,7 +503,8 @@ done:
     }
     p11prov_return_session(session);
     p11prov_obj_free(pub_key);
-    return priv_key;
+    *key = priv_key;
+    return ret;
 }
 
 static void p11prov_common_gen_cleanup(void *genctx)
@@ -612,12 +572,48 @@ static int p11prov_common_match(const void *keydata1, const void *keydata2,
 static void *p11prov_rsa_gen_init(void *provctx, int selection,
                                   const OSSL_PARAM params[])
 {
+    struct key_generator *ctx = NULL;
+    /* big endian 65537 */
+    unsigned char def_e[] = { 0x01, 0x00, 0x01 };
+    int ret;
+
     P11PROV_debug("rsa gen_init %p", provctx);
 
-    return p11prov_common_gen_init(provctx, selection, CKK_RSA, params);
+    ret = p11prov_ctx_status(provctx);
+    if (ret != CKR_OK) {
+        return NULL;
+    }
+
+    if ((selection & OSSL_KEYMGMT_SELECT_KEYPAIR) == 0) {
+        P11PROV_raise(provctx, CKR_ARGUMENTS_BAD, "Unsupported selection");
+        return NULL;
+    }
+
+    ctx = OPENSSL_zalloc(sizeof(struct key_generator));
+    if (ctx == NULL) {
+        P11PROV_raise(provctx, CKR_HOST_MEMORY, "Failed to get key_generator");
+        return NULL;
+    }
+    ctx->provctx = (P11PROV_CTX *)provctx;
+    ctx->type = CKK_RSA;
+
+    /* set defaults */
+    ctx->mechanism.mechanism = CKM_RSA_PKCS_KEY_PAIR_GEN;
+    ctx->data.rsa.modulus_bits = 2048;
+    ctx->data.rsa.exponent_size = sizeof(def_e);
+    memcpy(ctx->data.rsa.exponent, def_e, ctx->data.rsa.exponent_size);
+
+    ret = p11prov_common_gen_set_params(ctx, params);
+    if (ret != RET_OSSL_OK) {
+        p11prov_common_gen_cleanup(ctx);
+        ctx = NULL;
+    }
+    return ctx;
 }
 
-static void *p11prov_rsa_gen(void *genctx, OSSL_CALLBACK *cb_fn, void *cb_arg)
+static int p11prov_rsa_gen_internal(void *genctx, OSSL_CALLBACK *cb_fn,
+                                    void *cb_arg, void **key,
+                                    bool add_allow_mechs)
 {
     struct key_generator *ctx = (struct key_generator *)genctx;
 
@@ -634,25 +630,51 @@ static void *p11prov_rsa_gen(void *genctx, OSSL_CALLBACK *cb_fn, void *cb_arg)
           ctx->data.rsa.exponent_size },
     };
 #define RSA_PRIVKEY_TMPL_SIZE 6
-    CK_ATTRIBUTE privkey_template[RSA_PRIVKEY_TMPL_SIZE + COMMON_TMPL_SIZE] = {
+#define RSA_PRIVKEY_MAX RSA_PRIVKEY_TMPL_SIZE + 1 + COMMON_TMPL_SIZE
+    CK_ATTRIBUTE privkey_template[RSA_PRIVKEY_MAX] = {
         { CKA_TOKEN, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
         { CKA_PRIVATE, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
         { CKA_SENSITIVE, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
         { CKA_DECRYPT, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
         { CKA_SIGN, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
         { CKA_UNWRAP, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
+        /* 7. Optional CKA_ALLOWED_MECHANISMS */
         /* TODO?
          * CKA_SUBJECT
          * CKA_COPYABLE = true ?
          */
     };
+
     int pubtsize = RSA_PUBKEY_TMPL_SIZE;
     int privtsize = RSA_PRIVKEY_TMPL_SIZE;
 
+    if (add_allow_mechs) {
+        privkey_template[privtsize].type = CKA_ALLOWED_MECHANISMS;
+        privkey_template[privtsize].pValue =
+            DISCARD_CONST(ctx->data.rsa.allowed_types);
+        privkey_template[privtsize].ulValueLen =
+            ctx->data.rsa.allowed_types_size;
+        privtsize++;
+    }
+
     P11PROV_debug("rsa gen %p %p %p", genctx, cb_fn, cb_arg);
 
     return p11prov_common_gen(ctx, pubkey_template, privkey_template, pubtsize,
-                              privtsize, cb_fn, cb_arg);
+                              privtsize, cb_fn, cb_arg, key);
+}
+
+static void *p11prov_rsa_gen(void *genctx, OSSL_CALLBACK *cb_fn, void *cb_arg)
+{
+    struct key_generator *ctx = (struct key_generator *)genctx;
+    void *key;
+    CK_RV ret;
+
+    ret = p11prov_rsa_gen_internal(genctx, cb_fn, cb_arg, &key, false);
+    if (ret != CKR_OK) {
+        P11PROV_raise(ctx->provctx, ret, "RSA Key Gen failed");
+        return NULL;
+    }
+    return key;
 }
 
 static const OSSL_PARAM *p11prov_rsa_gen_settable_params(void *genctx,
@@ -681,7 +703,8 @@ static void *p11prov_rsa_new(void *provctx)
     }
 
     return p11prov_obj_new(provctx, CK_UNAVAILABLE_INFORMATION,
-                           CK_INVALID_HANDLE, CK_UNAVAILABLE_INFORMATION);
+                           CK_P11PROV_IMPORTED_HANDLE,
+                           CK_UNAVAILABLE_INFORMATION);
 }
 
 static void p11prov_rsa_free(void *key)
@@ -774,7 +797,7 @@ static int p11prov_rsa_export(void *keydata, int selection,
     P11PROV_CTX *ctx = p11prov_obj_get_prov_ctx(key);
     CK_OBJECT_CLASS class = p11prov_obj_get_class(key);
 
-    P11PROV_debug("rsa export %p", keydata);
+    P11PROV_debug("rsa export %p, selection= %d", keydata, selection);
 
     if (key == NULL) {
         return RET_OSSL_ERR;
@@ -786,7 +809,8 @@ static int p11prov_rsa_export(void *keydata, int selection,
 
     /* if anything else is asked for we can't provide it, so be strict */
     if ((class == CKO_PUBLIC_KEY) || (selection & ~(PUBLIC_PARAMS)) == 0) {
-        return p11prov_obj_export_public_key(key, CKK_RSA, true, cb_fn, cb_arg);
+        return p11prov_obj_export_public_key(key, CKK_RSA, true, false, cb_fn,
+                                             cb_arg);
     }
 
     return RET_OSSL_ERR;
@@ -988,22 +1012,14 @@ DISPATCH_KEYMGMT_FN(rsa, gen_settable_params);
 
 static CK_RV set_default_rsapss_mechanisms(struct key_generator *ctx)
 {
-    CK_MECHANISM_TYPE rsapss_mechs[] = {
-        CKM_SHA1_RSA_PKCS_PSS,     CKM_SHA224_RSA_PKCS_PSS,
-        CKM_SHA256_RSA_PKCS_PSS,   CKM_SHA384_RSA_PKCS_PSS,
-        CKM_SHA512_RSA_PKCS_PSS,   CKM_SHA3_224_RSA_PKCS_PSS,
-        CKM_SHA3_256_RSA_PKCS_PSS, CKM_SHA3_384_RSA_PKCS_PSS,
-        CKM_SHA3_512_RSA_PKCS_PSS
-    };
-
-    ctx->data.rsa.allowed_types = OPENSSL_malloc(sizeof(rsapss_mechs));
+    size_t mechs_size = sizeof(CK_MECHANISM_TYPE) * P11PROV_N_RSAPSS_MECHS;
+    ctx->data.rsa.allowed_types = OPENSSL_malloc(mechs_size);
     if (ctx->data.rsa.allowed_types == NULL) {
         P11PROV_raise(ctx->provctx, CKR_HOST_MEMORY, "Allocating data");
         return CKR_HOST_MEMORY;
     }
-    memcpy(ctx->data.rsa.allowed_types, rsapss_mechs, sizeof(rsapss_mechs));
-    ctx->data.rsa.allowed_types_size =
-        sizeof(rsapss_mechs) / sizeof(CK_MECHANISM_TYPE);
+    memcpy(ctx->data.rsa.allowed_types, p11prov_rsapss_mechs, mechs_size);
+    ctx->data.rsa.allowed_types_size = mechs_size;
 
     return CKR_OK;
 }
@@ -1013,63 +1029,149 @@ static void *p11prov_rsapss_gen(void *genctx, OSSL_CALLBACK *cb_fn,
 {
     struct key_generator *ctx = (struct key_generator *)genctx;
     CK_BBOOL token_supports_allowed_mechs = CK_TRUE;
-    P11PROV_OBJ *key = NULL;
+    bool set_quirk = false;
+    void *key = NULL;
     CK_RV ret;
 
-    key = p11prov_rsa_gen(genctx, cb_fn, cb_arg);
-    if (!key) {
-        return NULL;
-    }
-
-    /* params could already be caying pss restriction. If allowed_types
-     * is already set, skip setting defaults */
-    if (!ctx->data.rsa.allowed_types) {
-        ret = set_default_rsapss_mechanisms(ctx);
-        if (ret != CKR_OK) {
-            P11PROV_raise(ctx->provctx, ret, "Failed to get pss params");
-            goto done;
-        }
-    }
-
+    /* check if we can add CKA_ALLOWED_MECHANISMS at all */
     ret = p11prov_token_sup_attr(ctx->provctx, p11prov_obj_get_slotid(key),
                                  GET_ATTR, CKA_ALLOWED_MECHANISMS,
                                  &token_supports_allowed_mechs);
     if (ret != CKR_OK) {
-        P11PROV_raise(ctx->provctx, ret, "Failed to probe quirk");
+        P11PROV_raise(ctx->provctx, ret,
+                      "Failed to probe CKA_ALLOWED_MECHANISMS quirk");
         goto done;
     }
 
-    if (token_supports_allowed_mechs == CK_TRUE) {
-        CK_ATTRIBUTE template[] = {
-            { CKA_ALLOWED_MECHANISMS, ctx->data.rsa.allowed_types,
-              ctx->data.rsa.allowed_types_size },
-        };
-        CK_ULONG tsize = 1;
-
-        ret = p11prov_obj_set_attributes(ctx->provctx, NULL, key, template,
-                                         tsize);
-        if (ret != CKR_OK) {
-            P11PROV_debug("Failed to add RSAPSS restrictions (%lu)", ret);
-            if (ret == CKR_ATTRIBUTE_TYPE_INVALID) {
-                /* set quirk to disable future attempts for this token */
-                token_supports_allowed_mechs = CK_FALSE;
-                (void)p11prov_token_sup_attr(
-                    ctx->provctx, p11prov_obj_get_slotid(key), SET_ATTR,
-                    CKA_ALLOWED_MECHANISMS, &token_supports_allowed_mechs);
+    if (token_supports_allowed_mechs) {
+        /* We always want to restrict PSS keys to just the PSS mechanisms.
+         * If a specific restriction is not already set then set defaults */
+        if (!ctx->data.rsa.allowed_types) {
+            ret = set_default_rsapss_mechanisms(ctx);
+            if (ret != CKR_OK) {
+                P11PROV_raise(ctx->provctx, ret,
+                              "Failed to set default pss params");
+                goto done;
             }
         }
+
+        ret = p11prov_rsa_gen_internal(genctx, cb_fn, cb_arg, &key, true);
+        switch (ret) {
+        case CKR_OK:
+            break;
+        case CKR_ATTRIBUTE_TYPE_INVALID:
+            /* Failed: This may be because the token does not support
+             * CKA_ALLOWED_MECHANISMS, so we retry again later without it. */
+            P11PROV_debug("Failed to Generate PSS key with restrictions");
+
+            token_supports_allowed_mechs = CK_FALSE;
+            /* if the quirk has never been set we'll get back what we
+             * defaulted to, if that is the case then set the quirk.
+             * Otherwise the quirk was successfully probed earlier
+             * and we'll ignore setting anything as this may also be
+             * a fluke or a key specific failure */
+            ret = p11prov_token_sup_attr(
+                ctx->provctx, p11prov_obj_get_slotid(key), GET_ATTR,
+                CKA_ALLOWED_MECHANISMS, &token_supports_allowed_mechs);
+            if (ret != CKR_OK) {
+                P11PROV_raise(ctx->provctx, ret, "Failed to probe quirk");
+            } else if (token_supports_allowed_mechs == CK_FALSE) {
+                /* The previous check didn't hit a stored quirk, so
+                 * signal to set one later based on the outcome of the
+                 * next attempt. */
+                set_quirk = true;
+            }
+            break;
+        default:
+            /* In theory we should consider this error fatal, but given
+             * NSS sotoken returns CKR_GENERAL_ERROR just because it does
+             * not understand one attribute in the template we better
+             * retry anyway, as other tokens may return other errors too.
+             * The only penalty is that we won't set the quirk and keep
+             * retrying because we can't be sure the attribute is not
+             * valid in general ... */
+            break;
+        }
+    }
+
+    if (!key) {
+        ret = p11prov_rsa_gen_internal(genctx, cb_fn, cb_arg, &key, false);
+        if (ret != CKR_OK) {
+            goto done;
+        }
+
+        if (set_quirk) {
+            token_supports_allowed_mechs = CK_FALSE;
+            (void)p11prov_token_sup_attr(
+                ctx->provctx, p11prov_obj_get_slotid(key), SET_ATTR,
+                CKA_ALLOWED_MECHANISMS, &token_supports_allowed_mechs);
+        }
     }
 
     ret = CKR_OK;
 
 done:
     if (ret != CKR_OK) {
-        p11prov_obj_free(key);
-        key = NULL;
+        P11PROV_raise(ctx->provctx, ret, "Failed to generate RSA-PSS key");
+        return NULL;
     }
     return key;
 }
 
+static int p11prov_rsapss_gen_set_params(void *genctx,
+                                         const OSSL_PARAM params[])
+{
+    struct key_generator *ctx = (struct key_generator *)genctx;
+    const OSSL_PARAM *p;
+    int ret;
+
+    if (!ctx) {
+        return RET_OSSL_ERR;
+    }
+
+    if (params == NULL) {
+        return RET_OSSL_OK;
+    }
+
+    p = OSSL_PARAM_locate_const(params, OSSL_PKEY_PARAM_RSA_DIGEST);
+    if (p) {
+        CK_MECHANISM_TYPE digest_mech = CK_UNAVAILABLE_INFORMATION;
+        CK_MECHANISM_TYPE allowed_mech = CK_UNAVAILABLE_INFORMATION;
+        const char *digest = NULL;
+        CK_RV rv;
+
+        ret = OSSL_PARAM_get_utf8_string_ptr(p, &digest);
+        if (ret != RET_OSSL_OK) {
+            return ret;
+        }
+        rv = p11prov_digest_get_by_name(digest, &digest_mech);
+        if (rv != CKR_OK) {
+            ERR_raise(ERR_LIB_PROV, PROV_R_INVALID_DIGEST);
+            return RET_OSSL_ERR;
+        }
+        allowed_mech = p11prov_digest_to_rsapss_mech(digest_mech);
+        P11PROV_debug("Restrict RSAPSS DIGEST to %s (mech: %lu)", digest,
+                      allowed_mech);
+        if (allowed_mech == CK_UNAVAILABLE_INFORMATION) {
+            ERR_raise(ERR_LIB_PROV, PROV_R_INVALID_DIGEST);
+            return RET_OSSL_ERR;
+        }
+        /* overwrites any previous setting */
+        ctx->data.rsa.allowed_types_size = sizeof(CK_MECHANISM_TYPE);
+        ctx->data.rsa.allowed_types = OPENSSL_realloc(
+            ctx->data.rsa.allowed_types, ctx->data.rsa.allowed_types_size);
+
+        if (!ctx->data.rsa.allowed_types) {
+            ctx->data.rsa.allowed_types_size = 0;
+            ERR_raise(ERR_LIB_PROV, PROV_R_FAILED_TO_SET_PARAMETER);
+            return RET_OSSL_ERR;
+        }
+        ctx->data.rsa.allowed_types[0] = allowed_mech;
+    }
+
+    return p11prov_common_gen_set_params(genctx, params);
+}
+
 static const OSSL_PARAM *p11prov_rsapss_gen_settable_params(void *genctx,
                                                             void *provctx)
 {
@@ -1093,12 +1195,13 @@ const OSSL_DISPATCH p11prov_rsapss_keymgmt_functions[] = {
     DISPATCH_KEYMGMT_ELEM(rsa, NEW, new),
     DISPATCH_KEYMGMT_ELEM(rsa, GEN_INIT, gen_init),
     DISPATCH_KEYMGMT_ELEM(rsapss, GEN, gen),
-    DISPATCH_KEYMGMT_ELEM(common, GEN_SET_PARAMS, gen_set_params),
+    DISPATCH_KEYMGMT_ELEM(rsapss, GEN_SET_PARAMS, gen_set_params),
     DISPATCH_KEYMGMT_ELEM(rsapss, GEN_SETTABLE_PARAMS, gen_settable_params),
     DISPATCH_KEYMGMT_ELEM(common, GEN_CLEANUP, gen_cleanup),
     DISPATCH_KEYMGMT_ELEM(rsa, LOAD, load),
     DISPATCH_KEYMGMT_ELEM(rsa, FREE, free),
     DISPATCH_KEYMGMT_ELEM(rsa, HAS, has),
+    DISPATCH_KEYMGMT_ELEM(rsa, MATCH, match),
     DISPATCH_KEYMGMT_ELEM(rsa, IMPORT, import),
     DISPATCH_KEYMGMT_ELEM(rsa, IMPORT_TYPES, import_types),
     DISPATCH_KEYMGMT_ELEM(rsa, EXPORT, export),
@@ -1140,20 +1243,75 @@ static void *p11prov_ec_new(void *provctx)
     }
 
     return p11prov_obj_new(provctx, CK_UNAVAILABLE_INFORMATION,
-                           CK_INVALID_HANDLE, CK_UNAVAILABLE_INFORMATION);
+                           CK_P11PROV_IMPORTED_HANDLE,
+                           CK_UNAVAILABLE_INFORMATION);
 }
 
 static void *p11prov_ec_gen_init(void *provctx, int selection,
                                  const OSSL_PARAM params[])
 {
+    struct key_generator *ctx = NULL;
+    int ret;
+
     P11PROV_debug("ec gen_init %p", provctx);
 
-    return p11prov_common_gen_init(provctx, selection, CKK_EC, params);
+    ret = p11prov_ctx_status(provctx);
+    if (ret != CKR_OK) {
+        return NULL;
+    }
+
+    /* we need to allow to initialize a generation of just domain parameters,
+     * as this is used by OpenSSL for ECDH, to set the expected parameters
+     * first and then import the received public peer key */
+    if ((selection & OSSL_KEYMGMT_SELECT_ALL) == 0) {
+        P11PROV_raise(provctx, CKR_ARGUMENTS_BAD, "Unsupported selection");
+        return NULL;
+    }
+
+    ctx = OPENSSL_zalloc(sizeof(struct key_generator));
+    if (ctx == NULL) {
+        P11PROV_raise(provctx, CKR_HOST_MEMORY, "Failed to get key_generator");
+        return NULL;
+    }
+    ctx->provctx = (P11PROV_CTX *)provctx;
+    ctx->type = CKK_EC;
+
+    if ((selection & OSSL_KEYMGMT_SELECT_KEYPAIR) != 0) {
+        ctx->mechanism.mechanism = CKM_EC_KEY_PAIR_GEN;
+    } else {
+        ctx->mechanism.mechanism = CK_UNAVAILABLE_INFORMATION;
+    }
+
+    /* set defaults */
+    ctx->data.ec.ec_params = prime256v1_param;
+    ctx->data.ec.ec_params_size = sizeof(prime256v1_param);
+
+    ret = p11prov_common_gen_set_params(ctx, params);
+    if (ret != RET_OSSL_OK) {
+        p11prov_common_gen_cleanup(ctx);
+        ctx = NULL;
+    }
+    return ctx;
 }
 
 static void *p11prov_ec_gen(void *genctx, OSSL_CALLBACK *cb_fn, void *cb_arg)
 {
     struct key_generator *ctx = (struct key_generator *)genctx;
+    void *key;
+    CK_RV ret;
+
+    if (ctx->mechanism.mechanism == CK_UNAVAILABLE_INFORMATION) {
+        /* OpenSSL asked for a paramgen, basically it wants an
+         * empty key of a specific group that it will be filling
+         * up with public params later */
+        CK_ATTRIBUTE ec_params = {
+            .type = CKA_EC_PARAMS,
+            .pValue = (CK_VOID_PTR)ctx->data.ec.ec_params,
+            .ulValueLen = (CK_ULONG)ctx->data.ec.ec_params_size,
+        };
+
+        return mock_pub_ec_key(ctx->provctx, ctx->type, &ec_params);
+    }
 
     /* always leave space for CKA_ID and CKA_LABEL */
 #define EC_PUBKEY_TMPL_SIZE 5
@@ -1165,9 +1323,10 @@ static void *p11prov_ec_gen(void *genctx, OSSL_CALLBACK *cb_fn, void *cb_arg)
         { CKA_EC_PARAMS, (CK_BYTE *)ctx->data.ec.ec_params,
           ctx->data.ec.ec_params_size },
     };
-#define EC_PRIVKEY_TMPL_SIZE 5
+#define EC_PRIVKEY_TMPL_SIZE 6
     CK_ATTRIBUTE privkey_template[EC_PRIVKEY_TMPL_SIZE + COMMON_TMPL_SIZE] = {
         { CKA_TOKEN, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
+        { CKA_DERIVE, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
         { CKA_PRIVATE, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
         { CKA_SENSITIVE, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
         { CKA_SIGN, DISCARD_CONST(&val_true), sizeof(CK_BBOOL) },
@@ -1182,8 +1341,13 @@ static void *p11prov_ec_gen(void *genctx, OSSL_CALLBACK *cb_fn, void *cb_arg)
 
     P11PROV_debug("ec gen %p %p %p", ctx, cb_fn, cb_arg);
 
-    return p11prov_common_gen(ctx, pubkey_template, privkey_template, pubtsize,
-                              privtsize, cb_fn, cb_arg);
+    ret = p11prov_common_gen(ctx, pubkey_template, privkey_template, pubtsize,
+                             privtsize, cb_fn, cb_arg, &key);
+    if (ret != CKR_OK) {
+        P11PROV_raise(ctx->provctx, ret, "EC Key generation failed");
+        return NULL;
+    }
+    return key;
 }
 
 static const OSSL_PARAM *p11prov_ec_gen_settable_params(void *genctx,
@@ -1240,11 +1404,11 @@ static int p11prov_ec_match(const void *keydata1, const void *keydata2,
     return p11prov_common_match(keydata1, keydata2, CKK_EC, selection);
 }
 
-static int p11prov_ec_import(void *keydata, int selection,
-                             const OSSL_PARAM params[])
+static int p11prov_ec_import_genr(CK_KEY_TYPE key_type, void *keydata,
+                                  int selection, const OSSL_PARAM params[])
 {
     P11PROV_OBJ *key = (P11PROV_OBJ *)keydata;
-    CK_OBJECT_CLASS class = CKO_PUBLIC_KEY;
+    CK_OBJECT_CLASS class = CK_UNAVAILABLE_INFORMATION;
     CK_RV rv;
 
     P11PROV_debug("ec import %p", key);
@@ -1255,6 +1419,10 @@ static int p11prov_ec_import(void *keydata, int selection,
 
     if (selection & OSSL_KEYMGMT_SELECT_PRIVATE_KEY) {
         class = CKO_PRIVATE_KEY;
+    } else if (selection & OSSL_KEYMGMT_SELECT_PUBLIC_KEY) {
+        class = CKO_PUBLIC_KEY;
+    } else if (selection & OSSL_KEYMGMT_SELECT_DOMAIN_PARAMETERS) {
+        class = CKO_DOMAIN_PARAMETERS;
     }
 
     /* NOTE: the following is needed because of bug:
@@ -1270,19 +1438,32 @@ static int p11prov_ec_import(void *keydata, int selection,
         }
     }
 
-    rv = p11prov_obj_import_key(key, CKK_EC, class, params);
+    if (class == CKO_DOMAIN_PARAMETERS && key_type != CKK_EC) {
+        /* There are no Domain parameters associated with an
+         * ECX or RSA key in OpenSSL, so there is nothing really
+         * we can import */
+        return RET_OSSL_ERR;
+    }
+
+    rv = p11prov_obj_import_key(key, key_type, class, params);
     if (rv != CKR_OK) {
         return RET_OSSL_ERR;
     }
     return RET_OSSL_OK;
 }
 
+static int p11prov_ec_import(void *keydata, int selection,
+                             const OSSL_PARAM params[])
+{
+    return p11prov_ec_import_genr(CKK_EC, keydata, selection, params);
+}
+
 static int p11prov_ec_export(void *keydata, int selection, OSSL_CALLBACK *cb_fn,
                              void *cb_arg)
 {
     P11PROV_OBJ *key = (P11PROV_OBJ *)keydata;
     P11PROV_CTX *ctx = p11prov_obj_get_prov_ctx(key);
-    CK_OBJECT_CLASS class = p11prov_obj_get_class(key);
+    bool params_only = true;
 
     P11PROV_debug("ec export %p", keydata);
 
@@ -1290,16 +1471,21 @@ static int p11prov_ec_export(void *keydata, int selection, OSSL_CALLBACK *cb_fn,
         return RET_OSSL_ERR;
     }
 
+    if (selection & OSSL_KEYMGMT_SELECT_PRIVATE_KEY) {
+        /* can't export private keys */
+        return RET_OSSL_ERR;
+    }
+
     if (p11prov_ctx_allow_export(ctx) & DISALLOW_EXPORT_PUBLIC) {
         return RET_OSSL_ERR;
     }
 
-    /* this will return the public EC_POINT as well as DOMAIN_PARAMTERS */
-    if ((class == CKO_PUBLIC_KEY) || (selection & ~(PUBLIC_PARAMS)) == 0) {
-        return p11prov_obj_export_public_key(key, CKK_EC, true, cb_fn, cb_arg);
+    if (selection & OSSL_KEYMGMT_SELECT_PUBLIC_KEY) {
+        params_only = false;
     }
 
-    return RET_OSSL_ERR;
+    return p11prov_obj_export_public_key(key, CKK_EC, true, params_only, cb_fn,
+                                         cb_arg);
 }
 
 static const OSSL_PARAM p11prov_ec_key_types[] = {
@@ -1335,6 +1521,10 @@ static const char *p11prov_ec_query_operation_name(int operation_id)
     return NULL;
 }
 
+#define CURVE_521_BITS 521
+#define MAX_CURVE_BITS CURVE_521_BITS
+#define MAX_CURVE_SIZE ((MAX_CURVE_BITS + 7) / 8)
+
 static int p11prov_ec_secbits(int bits)
 {
     /* common values from various NIST documents */
@@ -1391,6 +1581,10 @@ static int p11prov_ec_get_params(void *keydata, OSSL_PARAM params[])
     if (p) {
         /* add room for ECDSA Signature DER overhead */
         CK_ULONG size = p11prov_obj_get_key_size(key);
+        if (size > MAX_CURVE_SIZE) {
+            /* coverity started looking for silly integer overflows */
+            return RET_OSSL_ERR;
+        }
         ret = OSSL_PARAM_set_int(p, 3 + (size + 4) * 2);
         if (ret != RET_OSSL_OK) {
             return ret;
@@ -1456,6 +1650,39 @@ static int p11prov_ec_get_params(void *keydata, OSSL_PARAM params[])
             p->data_size = pub_y->ulValueLen;
         }
     }
+    p = OSSL_PARAM_locate(params, OSSL_PKEY_PARAM_ENCODED_PUBLIC_KEY);
+    if (p) {
+        CK_ATTRIBUTE *pub_key;
+
+        if (p->data_type != OSSL_PARAM_OCTET_STRING) {
+            return RET_OSSL_ERR;
+        }
+
+        pub_key = p11prov_obj_get_ec_public_raw(key);
+        if (!pub_key) {
+            return RET_OSSL_ERR;
+        }
+
+        p->return_size = pub_key->ulValueLen;
+        if (p->data) {
+            if (p->data_size < pub_key->ulValueLen) {
+                return RET_OSSL_ERR;
+            }
+            memcpy(p->data, pub_key->pValue, pub_key->ulValueLen);
+        }
+    }
+    p = OSSL_PARAM_locate(params, OSSL_PKEY_PARAM_EC_POINT_CONVERSION_FORMAT);
+    if (p) {
+        bool compressed = p11prov_obj_get_ec_compressed(key);
+        if (compressed) {
+            ret = OSSL_PARAM_set_utf8_string(p, "compressed");
+        } else {
+            ret = OSSL_PARAM_set_utf8_string(p, "uncompressed");
+        }
+        if (ret != RET_OSSL_OK) {
+            return ret;
+        }
+    }
 
     return RET_OSSL_OK;
 }
@@ -1470,11 +1697,12 @@ static const OSSL_PARAM *p11prov_ec_gettable_params(void *provctx)
         OSSL_PARAM_utf8_string(OSSL_PKEY_PARAM_DEFAULT_DIGEST, NULL, 0),
         OSSL_PARAM_BN(OSSL_PKEY_PARAM_EC_PUB_X, NULL, 0),
         OSSL_PARAM_BN(OSSL_PKEY_PARAM_EC_PUB_Y, NULL, 0),
+        OSSL_PARAM_octet_string(OSSL_PKEY_PARAM_ENCODED_PUBLIC_KEY, NULL, 0),
+        OSSL_PARAM_utf8_string(OSSL_PKEY_PARAM_EC_POINT_CONVERSION_FORMAT, NULL,
+                               0),
         /*
-         * OSSL_PKEY_PARAM_ENCODED_PUBLIC_KEY
          * OSSL_PKEY_PARAM_EC_DECODED_FROM_EXPLICIT_PARAM
          * OSSL_PKEY_PARAM_EC_ENCODING
-         * OSSL_PKEY_PARAM_EC_POINT_CONVERSION_FORMAT
          * OSSL_PKEY_PARAM_EC_FIELD_TYPE
          * OSSL_PKEY_PARAM_EC_P
          * OSSL_PKEY_PARAM_EC_A
@@ -1575,8 +1803,35 @@ static void *p11prov_ed25519_gen_init(void *provctx, int selection,
 
     P11PROV_debug("ed25519 gen_init %p", provctx);
 
-    ctx = p11prov_common_gen_init(provctx, selection, CKK_EC_EDWARDS, curve);
-    if (!ctx) {
+    ret = p11prov_ctx_status(provctx);
+    if (ret != CKR_OK) {
+        return NULL;
+    }
+
+    /* we need to allow to initialize a generation of just domain parameters,
+     * as this is used by OpenSSL for ECDH, to set the expected parameters
+     * first and then import the received public peer key */
+    if ((selection & OSSL_KEYMGMT_SELECT_ALL) == 0) {
+        P11PROV_raise(provctx, CKR_ARGUMENTS_BAD, "Unsupported selection");
+        return NULL;
+    }
+
+    ctx = OPENSSL_zalloc(sizeof(struct key_generator));
+    if (ctx == NULL) {
+        P11PROV_raise(provctx, CKR_HOST_MEMORY, "Failed to get key_generator");
+        return NULL;
+    }
+    ctx->provctx = (P11PROV_CTX *)provctx;
+    ctx->type = CKK_EC_EDWARDS;
+
+    if ((selection & OSSL_KEYMGMT_SELECT_KEYPAIR) != 0) {
+        ctx->mechanism.mechanism = CKM_EC_EDWARDS_KEY_PAIR_GEN;
+    }
+
+    /* set defaults */
+    ret = p11prov_common_gen_set_params(ctx, curve);
+    if (ret != RET_OSSL_OK) {
+        p11prov_common_gen_cleanup(ctx);
         return NULL;
     }
 
@@ -1600,8 +1855,35 @@ static void *p11prov_ed448_gen_init(void *provctx, int selection,
 
     P11PROV_debug("ed448 gen_init %p", provctx);
 
-    ctx = p11prov_common_gen_init(provctx, selection, CKK_EC_EDWARDS, curve);
-    if (!ctx) {
+    ret = p11prov_ctx_status(provctx);
+    if (ret != CKR_OK) {
+        return NULL;
+    }
+
+    /* we need to allow to initialize a generation of just domain parameters,
+     * as this is used by OpenSSL for ECDH, to set the expected parameters
+     * first and then import the received public peer key */
+    if ((selection & OSSL_KEYMGMT_SELECT_ALL) == 0) {
+        P11PROV_raise(provctx, CKR_ARGUMENTS_BAD, "Unsupported selection");
+        return NULL;
+    }
+
+    ctx = OPENSSL_zalloc(sizeof(struct key_generator));
+    if (ctx == NULL) {
+        P11PROV_raise(provctx, CKR_HOST_MEMORY, "Failed to get key_generator");
+        return NULL;
+    }
+    ctx->provctx = (P11PROV_CTX *)provctx;
+    ctx->type = CKK_EC_EDWARDS;
+
+    if ((selection & OSSL_KEYMGMT_SELECT_KEYPAIR) != 0) {
+        ctx->mechanism.mechanism = CKM_EC_EDWARDS_KEY_PAIR_GEN;
+    }
+
+    /* set defaults */
+    ret = p11prov_common_gen_set_params(ctx, curve);
+    if (ret != RET_OSSL_OK) {
+        p11prov_common_gen_cleanup(ctx);
         return NULL;
     }
 
@@ -1642,7 +1924,7 @@ static int p11prov_ed_export(void *keydata, int selection, OSSL_CALLBACK *cb_fn,
 {
     P11PROV_OBJ *key = (P11PROV_OBJ *)keydata;
     P11PROV_CTX *ctx = p11prov_obj_get_prov_ctx(key);
-    CK_OBJECT_CLASS class = p11prov_obj_get_class(key);
+    bool params_only = true;
 
     P11PROV_debug("ed export %p", keydata);
 
@@ -1650,17 +1932,27 @@ static int p11prov_ed_export(void *keydata, int selection, OSSL_CALLBACK *cb_fn,
         return RET_OSSL_ERR;
     }
 
+    if (selection & OSSL_KEYMGMT_SELECT_PRIVATE_KEY) {
+        /* can't export private keys */
+        return RET_OSSL_ERR;
+    }
+
     if (p11prov_ctx_allow_export(ctx) & DISALLOW_EXPORT_PUBLIC) {
         return RET_OSSL_ERR;
     }
 
-    /* this will return the public EC_POINT */
-    if ((class == CKO_PUBLIC_KEY) || (selection & ~(PUBLIC_PARAMS)) == 0) {
-        return p11prov_obj_export_public_key(key, CKK_EC_EDWARDS, true, cb_fn,
-                                             cb_arg);
+    if (selection & OSSL_KEYMGMT_SELECT_PUBLIC_KEY) {
+        params_only = false;
     }
 
-    return RET_OSSL_ERR;
+    return p11prov_obj_export_public_key(key, CKK_EC_EDWARDS, true, params_only,
+                                         cb_fn, cb_arg);
+}
+
+static int p11prov_ed_import(void *keydata, int selection,
+                             const OSSL_PARAM params[])
+{
+    return p11prov_ec_import_genr(CKK_EC_EDWARDS, keydata, selection, params);
 }
 
 static const OSSL_PARAM *p11prov_ed_import_types(int selection)
@@ -1768,6 +2060,27 @@ static int p11prov_ed_get_params(void *keydata, OSSL_PARAM params[])
             return ret;
         }
     }
+    p = OSSL_PARAM_locate(params, OSSL_PKEY_PARAM_PUB_KEY);
+    if (p) {
+        CK_ATTRIBUTE *pub;
+
+        if (p->data_type != OSSL_PARAM_OCTET_STRING) {
+            return RET_OSSL_ERR;
+        }
+        ret = p11prov_obj_get_ed_pub_key(key, &pub);
+        if (ret != RET_OSSL_OK) {
+            return ret;
+        }
+
+        p->return_size = pub->ulValueLen;
+        if (p->data) {
+            if (p->data_size < pub->ulValueLen) {
+                return RET_OSSL_ERR;
+            }
+            memcpy(p->data, pub->pValue, pub->ulValueLen);
+            p->data_size = pub->ulValueLen;
+        }
+    }
 
     return RET_OSSL_OK;
 }
@@ -1775,7 +2088,7 @@ static int p11prov_ed_get_params(void *keydata, OSSL_PARAM params[])
 static const OSSL_PARAM *p11prov_ed_gettable_params(void *provctx)
 {
     static const OSSL_PARAM params[] = {
-        /* TODO: OSSL_PARAM_octet_string(OSSL_PKEY_PARAM_PUB_KEY, NULL, 0), */
+        OSSL_PARAM_octet_string(OSSL_PKEY_PARAM_PUB_KEY, NULL, 0),
         OSSL_PARAM_int(OSSL_PKEY_PARAM_BITS, NULL),
         OSSL_PARAM_int(OSSL_PKEY_PARAM_SECURITY_BITS, NULL),
         OSSL_PARAM_int(OSSL_PKEY_PARAM_MAX_SIZE, NULL),
@@ -1809,7 +2122,7 @@ const OSSL_DISPATCH p11prov_ed25519_keymgmt_functions[] = {
     DISPATCH_KEYMGMT_ELEM(ec, FREE, free),
     DISPATCH_KEYMGMT_ELEM(ec, HAS, has),
     DISPATCH_KEYMGMT_ELEM(ed, MATCH, match),
-    DISPATCH_KEYMGMT_ELEM(ec, IMPORT, import),
+    DISPATCH_KEYMGMT_ELEM(ed, IMPORT, import),
     DISPATCH_KEYMGMT_ELEM(ed, IMPORT_TYPES, import_types),
     DISPATCH_KEYMGMT_ELEM(ed, EXPORT, export),
     DISPATCH_KEYMGMT_ELEM(ed, EXPORT_TYPES, export_types),
@@ -1818,7 +2131,7 @@ const OSSL_DISPATCH p11prov_ed25519_keymgmt_functions[] = {
     DISPATCH_KEYMGMT_ELEM(ed, GETTABLE_PARAMS, gettable_params),
     DISPATCH_KEYMGMT_ELEM(ed, SET_PARAMS, set_params),
     DISPATCH_KEYMGMT_ELEM(ed, SETTABLE_PARAMS, settable_params),
-    /* TODO: match, validate, dup? */
+    /* TODO: validate, dup? */
     { 0, NULL },
 };
 
@@ -1833,7 +2146,7 @@ const OSSL_DISPATCH p11prov_ed448_keymgmt_functions[] = {
     DISPATCH_KEYMGMT_ELEM(ec, FREE, free),
     DISPATCH_KEYMGMT_ELEM(ec, HAS, has),
     DISPATCH_KEYMGMT_ELEM(ed, MATCH, match),
-    DISPATCH_KEYMGMT_ELEM(ec, IMPORT, import),
+    DISPATCH_KEYMGMT_ELEM(ed, IMPORT, import),
     DISPATCH_KEYMGMT_ELEM(ed, IMPORT_TYPES, import_types),
     DISPATCH_KEYMGMT_ELEM(ed, EXPORT, export),
     DISPATCH_KEYMGMT_ELEM(ed, EXPORT_TYPES, export_types),
@@ -1842,7 +2155,7 @@ const OSSL_DISPATCH p11prov_ed448_keymgmt_functions[] = {
     DISPATCH_KEYMGMT_ELEM(ed, GETTABLE_PARAMS, gettable_params),
     DISPATCH_KEYMGMT_ELEM(ed, SET_PARAMS, set_params),
     DISPATCH_KEYMGMT_ELEM(ed, SETTABLE_PARAMS, settable_params),
-    /* TODO: match, validate, dup? */
+    /* TODO: validate, dup? */
     { 0, NULL },
 };
 
