@@ -4,6 +4,8 @@
 #ifndef _STORE_H
 #define _STORE_H
 
+#include <openssl/core.h>
+
 #define DISPATCH_STORE_FN(name) DECL_DISPATCH_FUNC(store, p11prov_store, name)
 #define DISPATCH_STORE_ELEM(NAME, name) \
     { \
@@ -11,4 +13,8 @@
     }
 extern const OSSL_DISPATCH p11prov_store_functions[];
 
+int p11prov_store_direct_fetch(void *provctx, const char *uri,
+                               OSSL_CALLBACK *object_cb, void *object_cbarg,
+                               OSSL_PASSPHRASE_CALLBACK *pw_cb, void *pw_cbarg);
+
 #endif /* _STORE_H */
