@@ -59,9 +59,9 @@ void p11prov_uri_free(P11PROV_URI *parsed_uri);
 CK_OBJECT_CLASS p11prov_uri_get_class(P11PROV_URI *uri);
 void p11prov_uri_set_class(P11PROV_URI *uri, CK_OBJECT_CLASS class);
 CK_ATTRIBUTE p11prov_uri_get_id(P11PROV_URI *uri);
-void p11prov_uri_set_id(P11PROV_URI *uri, CK_ATTRIBUTE *id);
+CK_RV p11prov_uri_set_id(P11PROV_URI *uri, CK_ATTRIBUTE *id);
 CK_ATTRIBUTE p11prov_uri_get_label(P11PROV_URI *uri);
-void p11prov_uri_set_label(P11PROV_URI *uri, CK_ATTRIBUTE *label);
+CK_RV p11prov_uri_set_label(P11PROV_URI *uri, CK_ATTRIBUTE *label);
 char *p11prov_uri_get_serial(P11PROV_URI *uri);
 char *p11prov_uri_get_pin(P11PROV_URI *uri);
 CK_SLOT_ID p11prov_uri_get_slot_id(P11PROV_URI *uri);
@@ -120,7 +120,7 @@ static inline int constant_select_int(CK_ULONG cond, int a, int b)
     return (int)((A & mask) | (B & ~mask));
 }
 
-static inline void constant_select_buf(CK_ULONG cond, size_t size,
+static inline void constant_select_buf(CK_ULONG cond, CK_ULONG size,
                                        unsigned char *dst, unsigned char *a,
                                        unsigned char *b)
 {
@@ -132,4 +132,14 @@ static inline void constant_select_buf(CK_ULONG cond, size_t size,
     }
 }
 
+struct data_buffer {
+    uint8_t *data;
+    size_t length;
+};
+typedef struct data_buffer data_buffer;
+
+CK_RV p11prov_digest_util(P11PROV_CTX *provctx, const char *digest,
+                          const char *properties, data_buffer data[],
+                          data_buffer *output);
+
 #endif /* _UTIL_H */
