@@ -11,6 +11,41 @@ fi
 
 TOKENTYPE=$1
 
+# defaults -- overridden below or in the per-token setup
+SUPPORT_ED25519=1
+SUPPORT_ED448=1
+SUPPORT_RSA_PKCS1_ENCRYPTION=1
+SUPPORT_RSA_KEYGEN_PUBLIC_EXPONENT=1
+SUPPORT_TLSFUZZER=1
+SUPPORT_ALLOWED_MECHANISMS=0
+
+# Ed448 requires OpenSC 0.26.0
+OPENSC_VERSION=$(opensc-tool -i | grep OpenSC | sed -e "s/OpenSC 0\.\([0-9]*\).*/\1/")
+if [[ "$OPENSC_VERSION" -le "25" ]]; then
+    SUPPORT_ED448=0
+fi
+
+# FIPS Mode
+if [[ "${OPENSSL_FORCE_FIPS_MODE}" = "1" || "$(cat /proc/sys/crypto/fips_enabled)" = "1" ]]; then
+    # We can not use Edwards curves in FIPS mode
+    SUPPORT_ED25519=0
+    SUPPORT_ED448=0
+
+    # The FIPS does not allow the RSA-PKCS1.5 encryption
+    SUPPORT_RSA_PKCS1_ENCRYPTION=0
+
+    # The FIPS does not allow to set custom public exponent during key
+    # generation
+    SUPPORT_RSA_KEYGEN_PUBLIC_EXPONENT=0
+
+    # TLS Fuzzer does not work well in FIPS mode
+    SUPPORT_TLSFUZZER=0
+
+    # We also need additional configuration in openssl.cnf to assume the token
+    # is FIPS token
+    TOKENOPTIONS="pkcs11-module-assume-fips = true"
+fi
+
 # Temporary dir and Token data dir
 TMPPDIR="${TESTBLDDIR}/${TOKENTYPE}"
 TOKDIR="$TMPPDIR/tokens"
@@ -31,6 +66,8 @@ elif [ "${TOKENTYPE}" == "softokn" ]; then
     source "${TESTSSRCDIR}/softokn-init.sh"
 elif [ "${TOKENTYPE}" == "kryoptic" ]; then
     source "${TESTSSRCDIR}/kryoptic-init.sh"
+elif [ "${TOKENTYPE}" == "kryoptic.nss" ]; then
+    source "${TESTSSRCDIR}/kryoptic.nss-init.sh"
 else
     echo "Unknown token type: $1"
     exit 1
@@ -62,7 +99,7 @@ fi
 P11DEFARGS=("--module=${P11LIB}" "--login" "--pin=${PINVALUE}" "--token-label=${TOKENLABEL}")
 
 # prepare certtool configuration
-cat >> "${TMPPDIR}/cert.cfg" <<HEREDOC
+cat >> "${TMPPDIR}/cacert.cfg" <<HEREDOC
 ca
 cn = "Issuer"
 serial = 1
@@ -73,50 +110,80 @@ encryption_key
 cert_signing_key
 HEREDOC
 
+# Serial = 1 is the CA
+SERIAL=1
+
+crt_selfsign() {
+    LABEL=$1
+    CN=$2
+    KEYID=$3
+    ((SERIAL+=1))
+    sed -e "s|cn = .*|cn = $CN|g" \
+        -e "s|serial = .*|serial = $SERIAL|g" \
+        "${sed_inplace[@]}" "${TMPPDIR}/cacert.cfg"
+    "${certtool}" --generate-self-signed --outfile="${TMPPDIR}/${LABEL}.crt" \
+        --template="${TMPPDIR}/cacert.cfg" --provider="$P11LIB" \
+	--load-privkey "pkcs11:object=$LABEL;token=$TOKENLABELURI;type=private" \
+        --load-pubkey "pkcs11:object=$LABEL;token=$TOKENLABELURI;type=public" --outder 2>&1
+    pkcs11-tool "${P11DEFARGS[@]}" --write-object "${TMPPDIR}/${LABEL}.crt" --type=cert \
+        --id="$KEYID" --label="$LABEL" 2>&1
+}
+
 title LINE "Creating new Self Sign CA"
 KEYID='0000'
 URIKEYID="%00%00"
-CACRT="${TMPPDIR}/CAcert.crt"
-CACRT_PEM="${TMPPDIR}/CAcert.pem"
 CACRTN="caCert"
 pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="RSA:2048" \
 	--label="${CACRTN}" --id="${KEYID}" 2>&1
-"${certtool}" --generate-self-signed --outfile="${CACRT}" \
-	--template="${TMPPDIR}/cert.cfg" --provider="$P11LIB" \
-        --load-privkey "pkcs11:object=$CACRTN;token=$TOKENLABELURI;type=private" \
-        --load-pubkey "pkcs11:object=$CACRTN;token=$TOKENLABELURI;type=public" --outder 2>&1
-pkcs11-tool "${P11DEFARGS[@]}" --write-object "${CACRT}" --type=cert \
-        --id=$KEYID --label="$CACRTN" 2>&1
-
-# Serial = 1 is the CA
-SERIAL=2
+crt_selfsign $CACRTN "Issuer" $KEYID
 
 # convert the DER cert to PEM
+CACRT_PEM="${TMPPDIR}/${CACRTN}.pem"
+CACRT="${TMPPDIR}/${CACRTN}.crt"
 openssl x509 -inform DER -in "$CACRT" -outform PEM -out "$CACRT_PEM"
 
+CABASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
+CABASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
+CABASEURI="pkcs11:id=${URIKEYID}"
+CAPUBURI="pkcs11:type=public;id=${URIKEYID}"
+CAPRIURI="pkcs11:type=private;id=${URIKEYID}"
+CACRTURI="pkcs11:type=cert;object=${CACRTN}"
+
+title LINE "RSA PKCS11 URIS"
+echo "${CABASEURIWITHPINVALUE}"
+echo "${CABASEURIWITHPINSOURCE}"
+echo "${CABASEURI}"
+echo "${CAPUBURI}"
+echo "${CAPRIURI}"
+echo "${CACRTURI}"
+echo ""
+
+
+cat "${TMPPDIR}/cacert.cfg" > "${TMPPDIR}/cert.cfg"
 # the organization identification is not in the CA
 echo 'organization = "PKCS11 Provider"' >> "${TMPPDIR}/cert.cfg"
 # the cert_signing_key and "ca" should be only on the CA
 sed -e "/^cert_signing_key$/d" -e "/^ca$/d" "${sed_inplace[@]}" "${TMPPDIR}/cert.cfg"
 
 ca_sign() {
-    CRT=$1
-    LABEL=$2
-    CN=$3
-    KEYID=$4
+    LABEL=$1
+    CN=$2
+    KEYID=$3
+    shift 3
     ((SERIAL+=1))
     sed -e "s|cn = .*|cn = $CN|g" \
         -e "s|serial = .*|serial = $SERIAL|g" \
         -e "/^ca$/d" \
         "${sed_inplace[@]}" \
         "${TMPPDIR}/cert.cfg"
-    "${certtool}" --generate-certificate --outfile="${CRT}.crt" \
+    "${certtool}" --generate-certificate --outfile="${TMPPDIR}/${LABEL}.crt" \
         --template="${TMPPDIR}/cert.cfg" --provider="$P11LIB" \
 	--load-privkey "pkcs11:object=$LABEL;token=$TOKENLABELURI;type=private" \
         --load-pubkey "pkcs11:object=$LABEL;token=$TOKENLABELURI;type=public" --outder \
         --load-ca-certificate "${CACRT}" --inder \
-        --load-ca-privkey="pkcs11:object=$CACRTN;token=$TOKENLABELURI;type=private"
-    pkcs11-tool "${P11DEFARGS[@]}" --write-object "${CRT}.crt" --type=cert \
+        --load-ca-privkey="pkcs11:object=$CACRTN;token=$TOKENLABELURI;type=private" \
+        "$@"
+    pkcs11-tool "${P11DEFARGS[@]}" --write-object "${TMPPDIR}/${LABEL}.crt" --type=cert \
         --id="$KEYID" --label="$LABEL" 2>&1
 }
 
@@ -124,12 +191,11 @@ ca_sign() {
 # generate RSA key pair and self-signed certificate
 KEYID='0001'
 URIKEYID="%00%01"
-TSTCRT="${TMPPDIR}/testcert"
 TSTCRTN="testCert"
 
 pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="RSA:2048" \
 	--label="${TSTCRTN}" --id="$KEYID"
-ca_sign "$TSTCRT" $TSTCRTN "My Test Cert" $KEYID
+ca_sign "${TSTCRTN}" "My Test Cert" $KEYID
 
 BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -150,12 +216,11 @@ echo ""
 # generate ECC key pair
 KEYID='0002'
 URIKEYID="%00%02"
-ECCRT="${TMPPDIR}/eccert"
 ECCRTN="ecCert"
 
 pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="EC:secp256r1" \
 	--label="${ECCRTN}" --id="$KEYID"
-ca_sign "$ECCRT" $ECCRTN "My EC Cert" $KEYID
+ca_sign $ECCRTN "My EC Cert" $KEYID
 
 ECBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 ECBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -166,12 +231,11 @@ ECCRTURI="pkcs11:type=cert;object=${ECCRTN}"
 
 KEYID='0003'
 URIKEYID="%00%03"
-ECPEERCRT="${TMPPDIR}/ecpeercert"
 ECPEERCRTN="ecPeerCert"
 
 pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="EC:secp256r1" \
 	--label="$ECPEERCRTN" --id="$KEYID"
-ca_sign "$ECPEERCRT" $ECPEERCRTN "My Peer EC Cert" $KEYID
+crt_selfsign $ECPEERCRTN "My Peer EC Cert" $KEYID
 
 ECPEERBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 ECPEERBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -197,17 +261,15 @@ echo ""
 
 
 ## Softtokn does not support edwards curves yet
-if [ "${TOKENTYPE}" != "softokn" ]; then
-
+if [ "${SUPPORT_ED25519}" -eq 1 ]; then
     # generate ED25519
     KEYID='0004'
     URIKEYID="%00%04"
-    EDCRT="${TMPPDIR}/edcert"
     EDCRTN="edCert"
 
     pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="EC:edwards25519" \
     	--label="${EDCRTN}" --id="$KEYID"
-    ca_sign "$EDCRT" $EDCRTN "My ED25519 Cert" $KEYID
+    ca_sign $EDCRTN "My ED25519 Cert" $KEYID
 
     EDBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID};pin-value=${PINVALUE}"
     EDBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID};pin-source=file:${PINFILE}"
@@ -225,43 +287,40 @@ if [ "${TOKENTYPE}" != "softokn" ]; then
     echo "${EDCRTURI}"
 fi
 
-# FIXME The pkcs11-tool before OpenSC 0.26 does not support Ed448 so they can
-# not be generated here
-#
-# generate ED448
-#KEYID='0009'
-#URIKEYID="%00%09"
-#ED2CRT="${TMPPDIR}/ed2cert"
-#ED2CRTN="ed2Cert"
-#
-# pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="EC:edwards448" \
-# 	--label="${ED2CRTN}" --id="$KEYID"
-# ca_sign "$EDCRT" $ED2CRTN "My ED448 Cert" $KEYID
-#
-# ED2BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID};pin-value=${PINVALUE}"
-# ED2BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID};pin-source=file:${PINFILE}"
-# ED2BASEURI="pkcs11:id=${URIKEYID}"
-# ED2PUBURI="pkcs11:type=public;id=${URIKEYID}"
-# ED2PRIURI="pkcs11:type=private;id=${URIKEYID}"
-# ED2CRTURI="pkcs11:type=cert;object=${ED2CRTN}"
-#
-# title LINE "ED448 PKCS11 URIS"
-# echo "${EDBASEURIWITHPINVALUE}"
-# echo "${EDBASEURIWITHPINSOURCE}"
-# echo "${EDBASEURI}"
-# echo "${EDPUBURI}"
-# echo "${EDPRIURI}"
-# echo "${EDCRTURI}"
+if [ "${SUPPORT_ED448}" -eq 1 ]; then
+    # generate ED448
+    KEYID='0009'
+    URIKEYID="%00%09"
+    ED2CRTN="ed2Cert"
+
+    pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="EC:Ed448" \
+        --label="${ED2CRTN}" --id="$KEYID"
+    ca_sign $ED2CRTN "My ED448 Cert" $KEYID
+
+    ED2BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID};pin-value=${PINVALUE}"
+    ED2BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID};pin-source=file:${PINFILE}"
+    ED2BASEURI="pkcs11:id=${URIKEYID}"
+    ED2PUBURI="pkcs11:type=public;id=${URIKEYID}"
+    ED2PRIURI="pkcs11:type=private;id=${URIKEYID}"
+    ED2CRTURI="pkcs11:type=cert;object=${ED2CRTN}"
+
+    title LINE "ED448 PKCS11 URIS"
+    echo "${ED2BASEURIWITHPINVALUE}"
+    echo "${ED2BASEURIWITHPINSOURCE}"
+    echo "${ED2BASEURI}"
+    echo "${ED2PUBURI}"
+    echo "${ED2PRIURI}"
+    echo "${ED2CRTURI}"
+fi
 
 title PARA "generate RSA key pair, self-signed certificate, remove public key"
 KEYID='0005'
 URIKEYID="%00%05"
-TSTCRT="${TMPPDIR}/testcert2"
 TSTCRTN="testCert2"
 
 pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="RSA:2048" \
 	--label="${TSTCRTN}" --id="$KEYID"
-ca_sign "$TSTCRT" $TSTCRTN "My Test Cert 2" $KEYID
+ca_sign $TSTCRTN "My Test Cert 2" $KEYID
 pkcs11-tool "${P11DEFARGS[@]}" --delete-object --type pubkey --id 0005
 
 BASE2URIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
@@ -281,12 +340,11 @@ echo ""
 title PARA "generate EC key pair, self-signed certificate, remove public key"
 KEYID='0006'
 URIKEYID="%00%06"
-TSTCRT="${TMPPDIR}/eccert2"
 TSTCRTN="ecCert2"
 
 pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="EC:secp384r1" \
 	--label="${TSTCRTN}" --id="$KEYID"
-ca_sign "$TSTCRT" $TSTCRTN "My EC Cert 2" $KEYID
+ca_sign $TSTCRTN "My EC Cert 2" $KEYID
 pkcs11-tool "${P11DEFARGS[@]}" --delete-object --type pubkey --id 0006
 
 ECBASE2URIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
@@ -303,8 +361,8 @@ echo "${ECPRI2URI}"
 echo "${ECCRT2URI}"
 echo ""
 
-if [ -f /etc/redhat-release ]; then
-    title PARA "explicit EC unsupported on Fedora/EL"
+if [ -z "${ENABLE_EXPLICIT_EC_TEST}" ]; then
+    title PARA "explicit EC unsupported"
 elif [ "${TOKENTYPE}" == "softokn" ]; then
     title PARA "explicit EC unsupported with softokn"
 else
@@ -334,12 +392,11 @@ fi
 title PARA "generate EC key pair with ALWAYS AUTHENTICATE flag, self-signed certificate"
 KEYID='0008'
 URIKEYID="%00%08"
-TSTCRT="${TMPPDIR}/eccert3"
 TSTCRTN="ecCert3"
 
 pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="EC:secp521r1" \
 	--label="${TSTCRTN}" --id="$KEYID" --always-auth
-ca_sign "$TSTCRT" $TSTCRTN "My EC Cert 3" $KEYID
+ca_sign $TSTCRTN "My EC Cert 3" $KEYID
 
 ECBASE3URIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
 ECBASE3URIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
@@ -357,6 +414,63 @@ echo "${ECPRI3URI}"
 echo "${ECCRT3URI}"
 echo ""
 
+if [ "${SUPPORT_ALLOWED_MECHANISMS}" -eq 1 ]; then
+    # generate unrestricted RSA-PSS key pair and self-signed RSA-PSS certificate
+    KEYID='0010'
+    URIKEYID="%00%10"
+    TSTCRTN="testRsaPssCert"
+
+    pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="RSA:2048" \
+        --label="${TSTCRTN}" --id="$KEYID" --allowed-mechanisms \
+        RSA-PKCS-PSS,SHA1-RSA-PKCS-PSS,SHA224-RSA-PKCS-PSS,SHA256-RSA-PKCS-PSS,SHA384-RSA-PKCS-PSS,SHA512-RSA-PKCS-PSS
+    ca_sign "${TSTCRTN}" "My RsaPss Cert" $KEYID "--sign-params=RSA-PSS"
+
+    RSAPSSBASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
+    RSAPSSBASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
+    RSAPSSBASEURI="pkcs11:id=${URIKEYID}"
+    RSAPSSPUBURI="pkcs11:type=public;id=${URIKEYID}"
+    RSAPSSPRIURI="pkcs11:type=private;id=${URIKEYID}"
+    RSAPSSCRTURI="pkcs11:type=cert;object=${TSTCRTN}"
+
+    title LINE "RSA-PSS PKCS11 URIS"
+    echo "${RSAPSSBASEURIWITHPINVALUE}"
+    echo "${RSAPSSBASEURIWITHPINSOURCE}"
+    echo "${RSAPSSBASEURI}"
+    echo "${RSAPSSPUBURI}"
+    echo "${RSAPSSPRIURI}"
+    echo "${RSAPSSCRTURI}"
+    echo ""
+
+    # generate RSA-PSS (3k) key pair restricted to SHA256 digests
+    # and self-signed RSA-PSS certificate
+    KEYID='0011'
+    URIKEYID="%00%11"
+    TSTCRTN="testRsaPss2Cert"
+
+    pkcs11-tool "${P11DEFARGS[@]}" --keypairgen --key-type="RSA:3092" \
+        --label="${TSTCRTN}" --id="$KEYID" --allowed-mechanisms \
+        SHA256-RSA-PKCS-PSS
+    ca_sign "${TSTCRTN}" "My RsaPss2 Cert" $KEYID \
+        "--sign-params=RSA-PSS" "--hash=SHA256"
+
+    RSAPSS2BASEURIWITHPINVALUE="pkcs11:id=${URIKEYID}?pin-value=${PINVALUE}"
+    RSAPSS2BASEURIWITHPINSOURCE="pkcs11:id=${URIKEYID}?pin-source=file:${PINFILE}"
+    RSAPSS2BASEURI="pkcs11:id=${URIKEYID}"
+    RSAPSS2PUBURI="pkcs11:type=public;id=${URIKEYID}"
+    RSAPSS2PRIURI="pkcs11:type=private;id=${URIKEYID}"
+    RSAPSS2CRTURI="pkcs11:type=cert;object=${TSTCRTN}"
+
+    title LINE "RSA-PSS 2 PKCS11 URIS"
+    echo "${RSAPSS2BASEURIWITHPINVALUE}"
+    echo "${RSAPSS2BASEURIWITHPINSOURCE}"
+    echo "${RSAPSS2BASEURI}"
+    echo "${RSAPSS2PUBURI}"
+    echo "${RSAPSS2PRIURI}"
+    echo "${RSAPSS2CRTURI}"
+    echo ""
+fi
+
+
 title PARA "Show contents of ${TOKENTYPE} token"
 echo " ----------------------------------------------------------------------------------------------------"
 pkcs11-tool "${P11DEFARGS[@]}" -O
@@ -386,9 +500,14 @@ export OPENSSL_CONF="${OPENSSL_CONF}"
 export TESTSSRCDIR="${TESTSSRCDIR}"
 export TESTBLDDIR="${TESTBLDDIR}"
 
-export TESTPORT="${TESTPORT}"
+export SUPPORT_ED25519="${SUPPORT_ED25519}"
+export SUPPORT_ED448="${SUPPORT_ED448}"
+export SUPPORT_RSA_PKCS1_ENCRYPTION="${SUPPORT_RSA_PKCS1_ENCRYPTION}"
+export SUPPORT_RSA_KEYGEN_PUBLIC_EXPONENT="${SUPPORT_RSA_KEYGEN_PUBLIC_EXPONENT}"
+export SUPPORT_TLSFUZZER="${SUPPORT_TLSFUZZER}"
+export SUPPORT_ALLOWED_MECHANISMS="${SUPPORT_ALLOWED_MECHANISMS}"
 
-export CACRT="${CACRT_PEM}"
+export TESTPORT="${TESTPORT}"
 
 export TOKDIR="${TOKDIR}"
 export TMPPDIR="${TMPPDIR}"
@@ -396,6 +515,14 @@ export PINVALUE="${PINVALUE}"
 export SEEDFILE="${TMPPDIR}/noisefile.bin"
 export RAND64FILE="${TMPPDIR}/64krandom.bin"
 
+export CACRT="${CACRT_PEM}"
+export CABASEURIWITHPINVALUE="${CABASEURIWITHPINVALUE}"
+export CABASEURIWITHPINSOURCE="${CABASEURIWITHPINSOURCE}"
+export CABASEURI="${CABASEURI}"
+export CAPUBURI="${CAPUBURI}"
+export CAPRIURI="${CAPRIURI}"
+export CACRTURI="${CACRTURI}"
+
 export BASEURIWITHPINVALUE="${BASEURIWITHPINVALUE}"
 export BASEURIWITHPINSOURCE="${BASEURIWITHPINSOURCE}"
 export BASEURI="${BASEURI}"
@@ -449,6 +576,18 @@ export EDCRTURI="${EDCRTURI}"
 DBGSCRIPT
 fi
 
+if [ -n "${ED2BASEURI}" ]; then
+    cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
+
+export ED2BASEURIWITHPINVALUE="${ED2BASEURIWITHPINVALUE}"
+export ED2BASEURIWITHPINSOURCE="${ED2BASEURIWITHPINSOURCE}"
+export ED2BASEURI="${ED2BASEURI}"
+export ED2PUBURI="${ED2PUBURI}"
+export ED2PRIURI="${ED2PRIURI}"
+export ED2CRTURI="${ED2CRTURI}"
+DBGSCRIPT
+fi
+
 if [ -n "${ECXBASEURI}" ]; then
     cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
 
@@ -460,6 +599,25 @@ export ECXPRIURI="${ECXPRIURI}"
 DBGSCRIPT
 fi
 
+if [ -n "${RSAPSSBASEURI}" ]; then
+    cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
+
+export RSAPSSBASEURIWITHPINVALUE="${RSAPSSBASEURIWITHPINVALUE}"
+export RSAPSSBASEURIWITHPINSOURCE="${RSAPSSBASEURIWITHPINSOURCE}"
+export RSAPSSBASEURI="${RSAPSSBASEURI}"
+export RSAPSSPUBURI="${RSAPSSPUBURI}"
+export RSAPSSPRIURI="${RSAPSSPRIURI}"
+export RSAPSSCRTURI="${RSAPSSCRTURI}"
+
+export RSAPSS2BASEURIWITHPINVALUE="${RSAPSS2BASEURIWITHPINVALUE}"
+export RSAPSS2BASEURIWITHPINSOURCE="${RSAPSS2BASEURIWITHPINSOURCE}"
+export RSAPSS2BASEURI="${RSAPSS2BASEURI}"
+export RSAPSS2PUBURI="${RSAPSS2PUBURI}"
+export RSAPSS2PRIURI="${RSAPSS2PRIURI}"
+export RSAPSS2CRTURI="${RSAPSS2CRTURI}"
+DBGSCRIPT
+fi
+
 cat >> "${TMPPDIR}/testvars" <<DBGSCRIPT
 
 # for listing the separate pkcs11 calls
