@@ -9,26 +9,7 @@
 #include <openssl/evp.h>
 #include <openssl/store.h>
 #include <openssl/rand.h>
-
-static void hexify(char *out, unsigned char *byte, size_t len)
-{
-    char c[2], s;
-
-    for (size_t i = 0; i < len; i++) {
-        out[i * 3] = '%';
-        c[0] = byte[i] >> 4;
-        c[1] = byte[i] & 0x0f;
-        for (int j = 0; j < 2; j++) {
-            if (c[j] < 0x0A) {
-                s = '0';
-            } else {
-                s = 'a' - 10;
-            }
-            out[i * 3 + 1 + j] = c[j] + s;
-        }
-    }
-    out[len * 3] = '\0';
-}
+#include "util.h"
 
 static void check_rsa_key(EVP_PKEY *pubkey)
 {
@@ -37,7 +18,7 @@ static void check_rsa_key(EVP_PKEY *pubkey)
 
     ret = EVP_PKEY_get_bn_param(pubkey, OSSL_PKEY_PARAM_RSA_E, &tmp);
     if (ret != 1) {
-        fprintf(stderr, "Failed to get E param from public key");
+        PRINTERR("Failed to get E param from public key\n");
         exit(EXIT_FAILURE);
     } else {
         BN_free(tmp);
@@ -45,13 +26,13 @@ static void check_rsa_key(EVP_PKEY *pubkey)
     }
     ret = EVP_PKEY_get_bn_param(pubkey, OSSL_PKEY_PARAM_RSA_N, &tmp);
     if (ret != 1) {
-        fprintf(stderr, "Failed to get N param from public key");
+        PRINTERR("Failed to get N param from public key\n");
         exit(EXIT_FAILURE);
     } else {
         int bits;
         bits = EVP_PKEY_get_bits(pubkey);
         if (bits < 3072) {
-            fprintf(stderr, "Expected 3072 bits key, got %d\n", bits);
+            PRINTERR("Expected 3072 bits key, got %d\n", bits);
             exit(EXIT_FAILURE);
         }
         BN_free(tmp);
@@ -66,7 +47,7 @@ static void check_ec_key(EVP_PKEY *pubkey)
 
     ret = EVP_PKEY_get_bn_param(pubkey, OSSL_PKEY_PARAM_EC_PUB_X, &tmp);
     if (ret != 1) {
-        fprintf(stderr, "Failed to get X param from public key");
+        PRINTERR("Failed to get X param from public key\n");
         exit(EXIT_FAILURE);
     } else {
         BN_free(tmp);
@@ -74,7 +55,7 @@ static void check_ec_key(EVP_PKEY *pubkey)
     }
     ret = EVP_PKEY_get_bn_param(pubkey, OSSL_PKEY_PARAM_EC_PUB_Y, &tmp);
     if (ret != 1) {
-        fprintf(stderr, "Failed to get Y param from public key");
+        PRINTERR("Failed to get Y param from public key\n");
         exit(EXIT_FAILURE);
     } else {
         BN_free(tmp);
@@ -91,18 +72,18 @@ static void check_eddsa_key(EVP_PKEY *pubkey)
     ret = EVP_PKEY_get_octet_string_param(pubkey, OSSL_PKEY_PARAM_PUB_KEY, NULL,
                                           0, &len);
     if (ret != 1) {
-        fprintf(stderr, "Failed to get public key size");
+        PRINTERR("Failed to get public key size\n");
         exit(EXIT_FAILURE);
     }
     tmp = malloc(len);
     if (tmp == NULL) {
-        fprintf(stderr, "Failed to allocate memory for public key");
+        PRINTERR("Failed to allocate memory for public key\n");
         exit(EXIT_FAILURE);
     }
     ret = EVP_PKEY_get_octet_string_param(pubkey, OSSL_PKEY_PARAM_PUB_KEY, tmp,
                                           len, NULL);
     if (ret != 1) {
-        fprintf(stderr, "Failed to get public key");
+        PRINTERR("Failed to get public key\n");
         exit(EXIT_FAILURE);
     } else {
         free(tmp);
@@ -123,14 +104,14 @@ static void check_keys(OSSL_STORE_CTX *store, const char *key_type)
         switch (type) {
         case OSSL_STORE_INFO_PUBKEY:
             if (pubkey != NULL) {
-                fprintf(stderr, "Duplicate public key found!");
+                PRINTERR("Duplicate public key found!\n");
                 exit(EXIT_FAILURE);
             }
             pubkey = OSSL_STORE_INFO_get1_PUBKEY(info);
             break;
         case OSSL_STORE_INFO_PKEY:
             if (privkey != NULL) {
-                fprintf(stderr, "Duplicate private key found!");
+                PRINTERR("Duplicate private key found!\n");
                 exit(EXIT_FAILURE);
             }
             privkey = OSSL_STORE_INFO_get1_PKEY(info);
@@ -140,12 +121,12 @@ static void check_keys(OSSL_STORE_CTX *store, const char *key_type)
     }
 
     if (pubkey == NULL) {
-        fprintf(stderr, "Failed to load public key\n");
+        PRINTERR("Failed to load public key\n");
         exit(EXIT_FAILURE);
     }
 
     if (privkey == NULL) {
-        fprintf(stderr, "Failed to load private key\n");
+        PRINTERR("Failed to load private key\n");
         exit(EXIT_FAILURE);
     }
 
@@ -154,7 +135,8 @@ static void check_keys(OSSL_STORE_CTX *store, const char *key_type)
         check_rsa_key(pubkey);
     } else if (strcmp(key_type, "EC") == 0) {
         check_ec_key(pubkey);
-    } else if (strcmp(key_type, "ED25519") == 0) {
+    } else if (strcmp(key_type, "ED25519") == 0
+               || strcmp(key_type, "ED448") == 0) {
         check_eddsa_key(pubkey);
     }
 
@@ -176,26 +158,26 @@ static void gen_keys(const char *key_type, const char *label, const char *idhex,
 
     ctx = EVP_PKEY_CTX_new_from_name(NULL, key_type, "provider=pkcs11");
     if (ctx == NULL) {
-        fprintf(stderr, "Failed to init PKEY context\n");
+        PRINTERR("Failed to init PKEY context\n");
         exit(EXIT_FAILURE);
     }
 
     ret = EVP_PKEY_keygen_init(ctx);
     if (ret != 1) {
-        fprintf(stderr, "Failed to init keygen\n");
+        PRINTERR("Failed to init keygen\n");
         exit(EXIT_FAILURE);
     }
 
     ret = EVP_PKEY_CTX_set_params(ctx, params);
     if (ret != 1) {
-        fprintf(stderr, "Failed to set params\n");
+        PRINTERR("Failed to set params\n");
         exit(EXIT_FAILURE);
     }
 
     ret = EVP_PKEY_generate(ctx, &key);
     if (ret != 1) {
         if (!fail) {
-            fprintf(stderr, "Failed to generate key\n");
+            PRINTERR("Failed to generate key\n");
             exit(EXIT_FAILURE);
         }
         EVP_PKEY_CTX_free(ctx);
@@ -203,7 +185,7 @@ static void gen_keys(const char *key_type, const char *label, const char *idhex,
     }
 
     if (fail) {
-        fprintf(stderr, "Key generation unexpectedly succeeded\n");
+        PRINTERR("Key generation unexpectedly succeeded\n");
         exit(EXIT_FAILURE);
     }
 
@@ -211,7 +193,8 @@ static void gen_keys(const char *key_type, const char *label, const char *idhex,
         check_rsa_key(key);
     } else if (strcmp(key_type, "EC") == 0) {
         check_ec_key(key);
-    } else if (strcmp(key_type, "ED25519") == 0) {
+    } else if (strcmp(key_type, "ED25519") == 0
+               || strcmp(key_type, "ED448") == 0) {
         check_eddsa_key(key);
     }
 
@@ -225,13 +208,13 @@ static void gen_keys(const char *key_type, const char *label, const char *idhex,
 
     ret = asprintf(&uri, "pkcs11:id=%s", idhex);
     if (ret == -1) {
-        fprintf(stderr, "Failed to allocate uri\n");
+        PRINTERR("Failed to allocate uri\n");
         exit(EXIT_FAILURE);
     }
 
     store = OSSL_STORE_open(uri, NULL, NULL, NULL, NULL);
     if (store == NULL) {
-        fprintf(stderr, "Failed to open pkcs11 store\n");
+        PRINTERR("Failed to open pkcs11 store\n");
         exit(EXIT_FAILURE);
     }
     free(uri);
@@ -245,18 +228,18 @@ static void gen_keys(const char *key_type, const char *label, const char *idhex,
 
     store = OSSL_STORE_open("pkcs11:", NULL, NULL, NULL, NULL);
     if (store == NULL) {
-        fprintf(stderr, "Failed to open pkcs11 store\n");
+        PRINTERR("Failed to open pkcs11 store\n");
         exit(EXIT_FAILURE);
     }
 
     search = OSSL_STORE_SEARCH_by_alias(label);
     if (search == NULL) {
-        fprintf(stderr, "Failed to create store search filter\n");
+        PRINTERR("Failed to create store search filter\n");
         exit(EXIT_FAILURE);
     }
     ret = OSSL_STORE_find(store, search);
     if (ret != 1) {
-        fprintf(stderr, "Failed to set store search filter\n");
+        PRINTERR("Failed to set store search filter\n");
         exit(EXIT_FAILURE);
     }
     OSSL_STORE_SEARCH_free(search);
@@ -266,7 +249,8 @@ static void gen_keys(const char *key_type, const char *label, const char *idhex,
     OSSL_STORE_close(store);
 }
 
-static void sign_test(const char *label, bool fail)
+static void sign_test(const char *label, const char *mdname,
+                      const OSSL_PARAM *params, bool fail)
 {
     OSSL_STORE_CTX *store;
     OSSL_STORE_SEARCH *search;
@@ -279,20 +263,22 @@ static void sign_test(const char *label, bool fail)
     size_t siglen = 4096;
     int ret;
 
+    fprintf(stdout, "Test signature\n");
+
     store = OSSL_STORE_open("pkcs11:", NULL, NULL, NULL, NULL);
     if (store == NULL) {
-        fprintf(stderr, "Failed to open pkcs11 store\n");
+        PRINTERR("Failed to open pkcs11 store\n");
         exit(EXIT_FAILURE);
     }
 
     search = OSSL_STORE_SEARCH_by_alias(label);
     if (search == NULL) {
-        fprintf(stderr, "Failed to create store search filter\n");
+        PRINTERR("Failed to create store search filter\n");
         exit(EXIT_FAILURE);
     }
     ret = OSSL_STORE_find(store, search);
     if (ret != 1) {
-        fprintf(stderr, "Failed to set store search filter\n");
+        PRINTERR("Failed to set store search filter\n");
         exit(EXIT_FAILURE);
     }
     OSSL_STORE_SEARCH_free(search);
@@ -312,31 +298,32 @@ static void sign_test(const char *label, bool fail)
     OSSL_STORE_close(store);
 
     if (privkey == NULL) {
-        fprintf(stderr, "Failed to load private key\n");
+        PRINTERR("Failed to load private key\n");
         exit(EXIT_FAILURE);
     }
 
     ctx = EVP_MD_CTX_new();
     if (!ctx) {
-        fprintf(stderr, "Failed to init MD_CTX\n");
+        PRINTERR("Failed to init MD_CTX\n");
         exit(EXIT_FAILURE);
     }
 
-    ret = EVP_DigestSignInit(ctx, &pctx, EVP_sha256(), NULL, privkey);
+    ret =
+        EVP_DigestSignInit_ex(ctx, &pctx, mdname, NULL, NULL, privkey, params);
     if (ret == 0) {
-        fprintf(stderr, "Failed to init Sig Ctx\n");
+        PRINTERR("Failed to init Sig Ctx\n");
         exit(EXIT_FAILURE);
     }
 
     ret = EVP_DigestSign(ctx, sigret, &siglen, data, sizeof(data));
     if (ret == 0) {
         if (!fail) {
-            fprintf(stderr, "Failed to generate signature\n");
+            PRINTERR("Failed to generate signature\n");
             exit(EXIT_FAILURE);
         }
     } else {
         if (fail) {
-            fprintf(stderr, "Expected failure, but signature worked\n");
+            PRINTERR("Expected failure, but signature worked\n");
             exit(EXIT_FAILURE);
         }
     }
@@ -354,7 +341,7 @@ static char *tokenize(char **result, int max, char *str)
 
     copy = strdup(str);
     if (!copy) {
-        fprintf(stderr, "strdup failed\n");
+        PRINTERR("strdup failed\n");
         exit(EXIT_FAILURE);
     }
 
@@ -369,7 +356,7 @@ static char *tokenize(char **result, int max, char *str)
         }
         result[num] = strdup(tok);
         if (!result[num]) {
-            fprintf(stderr, "strdup failed\n");
+            PRINTERR("strdup failed\n");
             exit(EXIT_FAILURE);
         }
         num++;
@@ -411,19 +398,19 @@ int main(int argc, char *argv[])
         if (strcmp(tests[num], "RSA") == 0) {
             ret = RAND_bytes(id, 16);
             if (ret != 1) {
-                fprintf(stderr, "Failed to generate key id\n");
+                PRINTERR("Failed to generate key id\n");
                 exit(EXIT_FAILURE);
             }
             miniid = (id[0] << 24) + (id[1] << 16) + (id[2] << 8) + id[3];
             ret = asprintf(&label, "Test-RSA-gen-%08x", miniid);
             if (ret == -1) {
-                fprintf(stderr, "Failed to make label\n");
+                PRINTERR("Failed to make label\n");
                 exit(EXIT_FAILURE);
             }
             hexify(idhex, id, 16);
             ret = asprintf(&uri, "pkcs11:object=%s;id=%s", label, idhex);
             if (ret == -1) {
-                fprintf(stderr, "Failed to compose PKCS#11 URI\n");
+                PRINTERR("Failed to compose PKCS#11 URI\n");
                 exit(EXIT_FAILURE);
             }
             params[0] = OSSL_PARAM_construct_utf8_string("pkcs11_uri", uri, 0);
@@ -432,25 +419,28 @@ int main(int argc, char *argv[])
             params[2] = OSSL_PARAM_construct_end();
 
             gen_keys("RSA", label, idhex, params, false);
+
+            sign_test(label, "SHA256", NULL, false);
+
             free(label);
             free(uri);
 
         } else if (strcmp(tests[num], "RSA-PSS") == 0) {
             ret = RAND_bytes(id, 16);
             if (ret != 1) {
-                fprintf(stderr, "Failed to generate key id\n");
+                PRINTERR("Failed to generate key id\n");
                 exit(EXIT_FAILURE);
             }
             miniid = (id[0] << 24) + (id[1] << 16) + (id[2] << 8) + id[3];
             ret = asprintf(&label, "Test-RSA-PSS-gen-%08x", miniid);
             if (ret == -1) {
-                fprintf(stderr, "Failed to make label\n");
+                PRINTERR("Failed to make label\n");
                 exit(EXIT_FAILURE);
             }
             hexify(idhex, id, 16);
             ret = asprintf(&uri, "pkcs11:object=%s;id=%s", label, idhex);
             if (ret == -1) {
-                fprintf(stderr, "Failed to compose PKCS#11 URI\n");
+                PRINTERR("Failed to compose PKCS#11 URI\n");
                 exit(EXIT_FAILURE);
             }
             params[0] = OSSL_PARAM_construct_utf8_string("pkcs11_uri", uri, 0);
@@ -467,19 +457,19 @@ int main(int argc, char *argv[])
         } else if (strcmp(tests[num], "EC") == 0) {
             ret = RAND_bytes(id, 16);
             if (ret != 1) {
-                fprintf(stderr, "Failed to generate key id\n");
+                PRINTERR("Failed to generate key id\n");
                 exit(EXIT_FAILURE);
             }
             miniid = (id[0] << 24) + (id[1] << 16) + (id[2] << 8) + id[3];
             ret = asprintf(&label, "Test-EC-gen-%08x", miniid);
             if (ret == -1) {
-                fprintf(stderr, "Failed to make label\n");
+                PRINTERR("Failed to make label\n");
                 exit(EXIT_FAILURE);
             }
             hexify(idhex, id, 16);
             ret = asprintf(&uri, "pkcs11:object=%s;id=%s", label, idhex);
             if (ret == -1) {
-                fprintf(stderr, "Failed to compose PKCS#11 URI\n");
+                PRINTERR("Failed to compose PKCS#11 URI\n");
                 exit(EXIT_FAILURE);
             }
             params[0] = OSSL_PARAM_construct_utf8_string("pkcs11_uri", uri, 0);
@@ -488,25 +478,28 @@ int main(int argc, char *argv[])
             params[2] = OSSL_PARAM_construct_end();
 
             gen_keys("EC", label, idhex, params, false);
+
+            sign_test(label, "SHA256", NULL, false);
+
             free(label);
             free(uri);
 
         } else if (strcmp(tests[num], "RSAKeyUsage") == 0) {
             ret = RAND_bytes(id, 16);
             if (ret != 1) {
-                fprintf(stderr, "Failed to generate key id\n");
+                PRINTERR("Failed to generate key id\n");
                 exit(EXIT_FAILURE);
             }
             miniid = (id[0] << 24) + (id[1] << 16) + (id[2] << 8) + id[3];
             ret = asprintf(&label, "Test-RSA-Key-Usage-%08x", miniid);
             if (ret == -1) {
-                fprintf(stderr, "Failed to make label\n");
+                PRINTERR("Failed to make label\n");
                 exit(EXIT_FAILURE);
             }
             hexify(idhex, id, 16);
             ret = asprintf(&uri, "pkcs11:object=%s;id=%s", label, idhex);
             if (ret == -1) {
-                fprintf(stderr, "Failed to compose PKCS#11 URI\n");
+                PRINTERR("Failed to compose PKCS#11 URI\n");
                 exit(EXIT_FAILURE);
             }
             params[0] = OSSL_PARAM_construct_utf8_string("pkcs11_uri", uri, 0);
@@ -518,7 +511,7 @@ int main(int argc, char *argv[])
 
             gen_keys("RSA", label, idhex, params, false);
 
-            sign_test(label, true);
+            sign_test(label, "SHA256", NULL, true);
 
             params[1] = OSSL_PARAM_construct_utf8_string("pkcs11_key_usage",
                                                          (char *)bad_usage, 0);
@@ -527,38 +520,69 @@ int main(int argc, char *argv[])
 
             free(label);
             free(uri);
-        } else if (strcmp(tests[num], "ED25519") == 0) {
+        } else if (strcmp(tests[num], "ED25519") == 0
+                   || strcmp(tests[num], "ED448") == 0) {
+            const char *context = "context string";
+            const char *instance = "Ed25519ph";
+
+            if (strcmp(tests[num], "ED448") == 0) {
+                instance = "Ed448ph";
+            }
+
             ret = RAND_bytes(id, 16);
             if (ret != 1) {
-                fprintf(stderr, "Failed to generate key id\n");
+                PRINTERR("Failed to generate key id\n");
                 exit(EXIT_FAILURE);
             }
             miniid = (id[0] << 24) + (id[1] << 16) + (id[2] << 8) + id[3];
             ret = asprintf(&label, "Test-Ed-gen-%08x", miniid);
             if (ret == -1) {
-                fprintf(stderr, "Failed to make label\n");
+                PRINTERR("Failed to make label\n");
                 exit(EXIT_FAILURE);
             }
             hexify(idhex, id, 16);
             ret = asprintf(&uri, "pkcs11:object=%s;id=%s", label, idhex);
             if (ret == -1) {
-                fprintf(stderr, "Failed to compose PKCS#11 URI\n");
+                PRINTERR("Failed to compose PKCS#11 URI\n");
                 exit(EXIT_FAILURE);
             }
             params[0] = OSSL_PARAM_construct_utf8_string("pkcs11_uri", uri, 0);
             params[1] = OSSL_PARAM_construct_end();
 
-            gen_keys("ED25519", label, idhex, params, false);
+            gen_keys(tests[num], label, idhex, params, false);
+
+            sign_test(label, NULL, NULL, false);
+
+/* these are not defined in OpenSSL 3.0 so just skip the test */
+#ifdef OSSL_SIGNATURE_PARAM_CONTEXT_STRING
+            /* Test again with context string */
+            params[0] = OSSL_PARAM_construct_octet_string(
+                OSSL_SIGNATURE_PARAM_CONTEXT_STRING, (void *)context,
+                sizeof(context));
+            params[1] = OSSL_PARAM_construct_end();
+            sign_test(label, NULL, params, false);
+
+            /* Test again with prehash */
+            params[0] = OSSL_PARAM_construct_utf8_string(
+                OSSL_SIGNATURE_PARAM_INSTANCE, (char *)instance,
+                strlen(instance));
+            params[1] = OSSL_PARAM_construct_end();
+            sign_test(label, NULL, params, false);
+#else
+            (void)instance;
+            (void)context;
+#endif
+
             free(label);
             free(uri);
         } else {
-            fprintf(stderr, "Unknown test type [%s]\n", tests[num]);
+            PRINTERR("Unknown test type [%s]\n", tests[num]);
             exit(EXIT_FAILURE);
         }
     }
 
     freetokens(tests);
     free(copy);
-    fprintf(stderr, "Performed tests: %d\n", num);
+    PRINTERR("Performed tests: %d\n", num);
     exit(EXIT_SUCCESS);
 }
